/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright by The HDF Group.                                               *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of the HDF NPP Aggregation Tool (nagg). The full        *
 * NPP Aggregation Tool copyright notice, including terms governing use,     *
 * modification, and redistribution, is contained in the file COPYING,       *
 * which can be found at the root of the source code distribution tree.      *
 * If you do not have access to this file, you may request a copy from       *
 * help@hdfgroup.org.                                                        *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* This contains all routines involve in parse the command options. */

#include <nagg.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

/* local macro definitions */
#if 0
/* disable this definition since it is not needed now. But keep it for */
/* possible future use if input file can be a dir. */
#define NAGG_inputfiles_Max	100	/* max number of input files */
#endif

/* local variables */
static options_t nagg_options;		/* holds the options of the Nagg tool */
static options_p_t nagg_options_p=&nagg_options;
static char	like_products_list_array[NAGG_Product_list_max][DPID_size+1];
char	*like_products_list[NAGG_Product_list_max];
int      like_nproducts=0;
int      like_ngranules=-1;
int      like_ispackaged=0;

/* local prototypes */
static void dprint_products_list(int nlist, char **plist);
static int get_example_props(char *exfile, int *ngranules, int *nproducts,
    char **products_list, int *ispackaged);
static int get_product_props (hid_t loc_id, const char *name, 
    const H5L_info_t *info, void *operator_data);
static int set_geoproduct_to_process(char **products_list);
static int compatible_products(char *next_geoproduct, char *chkgeoproduct);

/* Function definitions */

/* Display the usage page of the nagg tool.
 * Return: no return value.
 */
void
usage(void)
{
    fprintf(stderr,
"usage:\n"
"%s [OPTIONS] <input_file>...\n"
"OPTIONS\n"
"    -h, --help\n"
"       Print command syntax; also list valid types and packaging groupings.\n"
"    --version\n"
"       Print version information.\n"
"    -t <list>, --type=<list>\n"
"       <list> specifies a comma separated list of NPP record type mnemonics.\n"
"       Unless -S is specified, the granule types will be packaged together.\n"
"       Types must be compatible to be packaged together. (Use -h to list\n"
"       valid package groupings). If \-t is not specified, -g <list> must be\n"
"       given directly or through -l <example> to aggregate only Geolocation\n"
"       product granules.\n"
"    -n <N>, --number=<N>\n"
"       <N> is the number of granules of each product in each aggregate file;\n"
"       must be greater than zero. If not specified, default is 1 (single\n"
"       granule files).\n"
"    -A <t>\n"
"       <t> is the number of seconds of aggregation time span of each\n"
"       aggregate file; must be a positive integer. It is exclusive with the\n"
"       -n option.\n"
"    -g <criterion>, --geolocation=<criterion>\n"
"       <criterion> is the criterion for searching the Geolocation granules\n"
"          no  | 0: Aggregate product files without Geolocation input or output.\n"
"          yes | 1: (Default) Allow approximate matching of Geolocation input\n"
"                   filenames.\n"
"       strict | 2: Require exact matching of Geolocation input filenames.\n"
"       <geoproduct> - With no data product specified: aggregate Geolocation\n"
"                      product <geoproduct> only.\n"
"                    - With data products: aggregate specified Geolocation\n" 
"                      product <geoproduct> in addition to the data products.\n" 
"    -l <example>, --like=<example>\n"
"       Package like the <example> file in number of granules, type list, and\n"
"       Geolocation product.  Example file with /N_GEO_Ref attribute will\n"
"       trigger -S option.\n"
"       <example> must be an existing NPP file. Options on the command line\n"
"       override the number of granules, type list and geolocation options\n"
"       from the <example> file.\n"
"    -d <directory>, --directory=<directory>\n"
"       <directory> is name of an existing directory in which the output NPP\n"
"       files are created.  Default is to create output files in the current\n"
"       directory.\n"
"    -S, --simple\n"
"       Simple aggregates are produced. Each type is packaged separately.\n"
"       Default is all types are packaged in a file.\n"
"    -O <origin>\n"
"       <origin> is the origin identifier, a four character string.\n"
"       Default is \"XXXX\".\n"
"    -D <domain>\n"
"       <domain> is the domain identifier, a three character string.\n"
"       Default is \"XXX\".\n"
"    --debug\n"
"       Read in all granules in the input files, including those not specified\n"
"       by the -t list. Display all the granules and end the execution without\n"
"       generating the normal output files.\n"
"       Note that this option is intended for tool debugging. Files or output\n"
"       generated should not be used for production purpose. Its behavior may\n"
"       change from version to version.\n"
"\n"
"    <input_file> is a list of one or more NPP files\n"
"\n"
    , progname_g);
}


/* functions to manage the Options Structure */

/* Initialize the option structure.
 * Return 0 if succeed, otherwise -1.
 */
int
init_option(const options_p_t options)
{
    if (options){
	HDmemset(options, 0, sizeof(options_t));
	return 0;
    }else
	return -1;
}

/* Retrieve the value of number of granules.
 * If options is NULL, return the ngranules of the tool options;
 * otherwise, return the ngranules of the given options.
 */
int
getopt_ngranules(const options_p_t options)
{
    if (options)
	return((*options).ngranulesperfile);
    else
	return(nagg_options.ngranulesperfile);
};

/* Set the values of number of granules to ngranules.
 * If options is NULL, set the value of the ngranules in the tool options;
 * otherwise, set the one in the given options.
 */
void
setopt_ngranules(const options_p_t options, const int ngranules)
{
    if (options)
	(*options).ngranulesperfile = ngranules;
    else
	nagg_options.ngranulesperfile = ngranules;
};

/* Retrieve the products list.
 * If options is NULL, return the products list and nproducts of the tool options
 * tthrough the pointers of products_list and nproducts.
 * Otherwise, return the values pair of the given options.
 */
void
getopt_products_list(const options_p_t options, char **products_list, int *nproducts)
{
    if (options){
	*products_list = &((*options).products_list_array);
	*nproducts = (*options).nproducts;
    }else{
	*products_list = &(nagg_options.products_list_array);
	*nproducts = nagg_options.nproducts;
    }
};

/* Set the values of products list and number of products to products_list
 * and nproducts.
 * If options is NULL, set the products list and nproducts in the tool options;
 * otherwise, set the ones in the given options.
 */
void
setopt_products_list(const options_p_t options, const char** products_list,
    const int nproducts)
{
    if (options){
	/* (*options).products_list = products_list; */
	(*options).nproducts = nproducts;
    }else{
	/* nagg_options.products_list = products_list; */
	nagg_options.nproducts = nproducts;
    }
}


/* Purpose:
 *    Retrieve the output directory argument.
 * return: output directory given by the -d option;
 *    null if not given.
 */
const char *
get_outdir(void)
{
    return(outDir);
}

/* Purpose:
 *    Print version information.
 *
 * Return:     Success:     0
 *             Failure:     -1
 */
int
print_nagg_version()
{
    int      status = 0;
    int      haszlib = 0;
    int      hasszlib = 0;
    int      hasencoder = 0;
    unsigned majnum, minnum, relnum;

    HDfprintf(stdout, "\n%s: Version %d.%d.%d-%s, %s\n", progname_g, NAGG_VER_MAJOR, NAGG_VER_MINOR, NAGG_VER_RELEASE, NAGG_VER_SUBRELEASE, NAGG_RELEASE_DATE);
    HDfprintf(stdout, "\nBuilt on these libraries:\n\n");
    HDfprintf(stdout, "HDF5_HL_REGION");
    /* TODO:  print HDF5_HL_REGION version */
    HDfprintf(stdout, ", using external libraries:\n");
    HDfprintf(stdout, "     hdf5lib ");
    /* TODO: print version of hdf5lib used to build HDF5_HL_REGION */
    HDfprintf(stdout, "\n\n");
     
     
    status = H5get_libversion(&majnum, &minnum, &relnum);
    if (status < 0) {
        HDfprintf(stderr, "Error:  Failed to get hdf5 library version.\n");
    } else {
        HDfprintf(stdout, "HDF5 library version: %d.%d.%d", majnum, minnum, relnum);
    }
    if(H5Zfilter_avail(H5Z_FILTER_DEFLATE)) {
        haszlib = 1;
    }
    if(H5Zfilter_avail(H5Z_FILTER_SZIP)) {
        hasszlib = 1;
    }
    if (haszlib || hasszlib) {
        HDfprintf(stdout, ", using external libraries:");
    }
    HDfprintf(stdout, "\n");
    
    if (haszlib) {
        HDfprintf(stdout, "     deflate(zlib)");
        /* TODO: print zlib version */
        HDfprintf(stdout, "\n");
    }
    if (hasszlib) {    
        unsigned int filterconfig;
        status = H5Zget_filter_info(H5Z_FILTER_SZIP, &filterconfig);
        if (status < 0) {
            HDfprintf(stderr, "Error:  Failed to get filter info for szip.\n");
        } else {
            HDfprintf(stdout, "     szip");
            if(H5Z_FILTER_CONFIG_ENCODE_ENABLED & filterconfig) {
                HDfprintf(stdout, "(encoder)");
            } else {
                HDfprintf(stdout, "(decoder)");
            }
            /* TODO:  print szip version */
        }
        HDfprintf(stdout, "\n");
    }
    HDfprintf(stdout, "\n");
    return status;
}


/* Private functions */

/* Purpose:
 *    Verify if a directory already exists.
 * Return TRUE if it is an existing directory;
 *        FALSE otherwise.
 */
static int
isdirectory(char *name){
    stat_t	st;

    /* sanity check */
    if (!name)
	return(FALSE);
    if (HDstat(name, &st)<0)
	return(FALSE);
    if (S_ISDIR(st.st_mode))
	return TRUE;
    else
	return FALSE;
}

typedef struct {
    char   *filepath;
    hid_t   grpid;
    char  **products_list;
    int    *nproducts;
    int    *ngranules;
    char  *geoproduct;
} ExFile_data;

/*
 * Function:  get_example_props
 *
 * Purpose:   Retrieve NPP file properties from the example files.
 *            Properties retrieved are: number of granules, products list.
 *
 * Parameter: IN exfile: example file name
 *            OUT ngranules: return number of granules per product
 *            OUT nproducts: return number of products
 *            OUT products_list: return the list of products (up to
 *            NAGG_Product_list_max products)

 *
 * Procedure:  Open filename (example file for output files)
 *             Get product names and aggregateNumberGranules from group_finder
 *             or a similar functor for an iterator over subgroups of
 *             /Data_Products.
 *
 * Structure:  We can iterate through the /Data_Products group to find the products
 *             and the aggregate number in the example file.  We need a structure to
 *             pass in parameters and retrieve the information.  It will be similar
 *             to GranF_data, but probably enough different to warrant a completely
 *             different struct.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             August 22, 2012
 */

int
get_example_props(char *exfile, int *ngranules, int *nproducts,
                  char **products_list, int *ispackaged)
{
    int         status = SUCCEED;
    ExFile_data exf_dat;
    hid_t       file = -1;       /* File and group identifiers */
    hid_t       grp = -1;
    htri_t      retval;
    char *like_geoproduct;


    if (NULL == (like_geoproduct = (char *)HDcalloc(DPID_size, sizeof(char)))) {
        NAGG_ERROR("parse_options():  HDcalloc failed\n");
        leave(EXIT_FAILURE);
    }


    if((file = H5Fopen(exfile, H5F_ACC_RDONLY, H5P_DEFAULT)) < 0) {
        retval = H5Fis_hdf5(exfile);
        if (retval > 0) {
            NAGG_ERROR("get_example_props():\n");
            HDfprintf(stderr, "%s is an HDF5 file, but H5Fopen failed.\n", exfile);
        } else if (!retval) {
            NAGG_ERROR("get_example_props():\n");
            HDfprintf(stderr, "%s is not an HDF5 file.\n", exfile);
        } else {
            NAGG_ERROR("get_example_props():\n");
            HDfprintf(stderr, "%s does not exist or otherwise cannot be accessed for reading.\n",
            exfile);
        }
        status = FAIL;
        goto done;
    }

    /*
     * Open the /Data_Products group,
     */
    if((grp = H5Gopen2(file, "/Data_Products", H5P_DEFAULT)) < 0) {
        NAGG_ERROR("get_example_props():\n");
        HDfprintf(stderr, "Failed to open /Data_Products group in example file %s.\n", exfile);
        status = FAIL;
        goto done;
    }
    /* Fill in values in struct for iterating through product groups */
    exf_dat.filepath = exfile;
    exf_dat.grpid = grp;
    exf_dat.products_list = products_list;
    exf_dat.nproducts = nproducts;
    exf_dat.ngranules = ngranules;
    exf_dat.geoproduct = like_geoproduct;
    if ( H5Aexists_by_name(file, "/", "N_GEO_Ref", H5P_DEFAULT))
        *ispackaged = 0;
    else
        *ispackaged = 1;

    /* Iterate over the objects in the /Data_Products group to get the products and the aggregation number */
    if((status = H5Literate(grp, H5_INDEX_NAME, H5_ITER_NATIVE, NULL, get_product_props, &exf_dat)) < 0) {
        NAGG_ERROR("get_example_props(): Failed to get example file info iterating over product groups.\n");
        goto done;
    }
    /* If we're finding a geoproduct from the example file, it seems it is  
       necessary to allocate memory for the DPID or it gets lost when 
       exiting the function.  However, "" shouldn't be assigned to 
       geoproduct. */ 
    if (!geoproduct && HDstrneq("", exf_dat.geoproduct))
        geoproduct = exf_dat.geoproduct;
done:
    if (grp > -1)
        H5Gclose(grp);
    if (file > -1)
        H5Fclose(file);
    return status;
}

/*
 * Function:   get_product_props
 *
 * Purpose:    Function for iterator used to iterate over the subgroups of
 *             the /Data_Products group in the nagg example file in order to
 *             find the products and the aggregation number.
 *
 * Note:       There will be one subgroup for each data product in a packaged
 *             file.  The subgroups have one dataset with a name ending in _Aggr
 *             with attributes describing the aggregation, one being
 *             AggregateNumberGranules.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             August 22, 2012
 *
 */
static int
get_product_props (hid_t loc_id, const char *name, const H5L_info_t *info,
                       void *operator_data)
{
    int        status = 0;
    H5O_info_t infobuf;
    hid_t      grpid = -1;  /* group identifier */
    hid_t      dset = -1;
    hid_t      attr = -1;
    char       linkName[NAGG_Product_Path_size + 1];
    char      *dpid;
    
    int aggr_num_granules;
    int number_products;

    ExFile_data *exf_dat = (ExFile_data *)operator_data;
    number_products = *exf_dat->nproducts;

    if((status = H5Oget_info_by_name (loc_id, name, &infobuf, H5P_DEFAULT)) < 0) {
        NAGG_ERROR("get_product_props():\n");
        HDfprintf(stderr, "Failed to get info for %s.\n", name);
        goto done;
    }
   switch (infobuf.type) {
        case H5O_TYPE_GROUP:
            if((grpid = H5Gopen2(loc_id, name, H5P_DEFAULT))<0) {
                NAGG_ERROR("get_product_props():\n");
                HDfprintf(stderr, "Failed to open product group %s in file %s.\n", linkName,
                        exf_dat->filepath);
                status = FAIL;
                goto done;
            }
            
            dpid = get_product_id_by_sname(name);
            if(*dpid == 'G' && NULL == get_gpid_by_id(dpid) ) {
                /* product is GEO product.  Assign dpid to geoproduct.  */
		HDstrcpy(exf_dat->geoproduct, dpid);
            } else { 
                /* the group name matches the product name. */
                HDstrcpy(exf_dat->products_list[number_products], dpid);

                /* increment the number of products. */
                *exf_dat->nproducts = ++number_products;
            }

            /* get the number of granules in the aggregation from the <CollectionShortName>_Aggr dataset. */
            snprintf(linkName, HDstrlen(name) + 6,  "%s_Aggr", name);
            if((dset = H5Dopen2(grpid, linkName, H5P_DEFAULT))<0) {
                NAGG_ERROR("get_product_props():\n");
                HDfprintf(stderr, "Failed to open dataset %s in file %s.\n", linkName,
                        exf_dat->filepath);
                status = FAIL;
                goto done;
            }
            if((attr = H5Aopen(dset, "AggregateNumberGranules", H5P_DEFAULT)) < 0) {
                NAGG_ERROR("get_product_props():\n");
                HDfprintf(stderr, "Failed to open attribute AggregateNumberGranules on dataset %s.\n", linkName);
                status = FAIL;
                goto done;
            } else if((status = H5Aread(attr, H5T_NATIVE_INT, &aggr_num_granules)) < 0) {
                NAGG_ERROR("get_product_props():\n");
                HDfprintf(stderr, "Failed to read attribute AggregateNumberGranules on dataset %s.\n", name);
                status = FAIL;
                goto done;
            } else if(*exf_dat->nproducts > 1) {
                /* If there are multiple products in the example file, the number of granules
                   for this product must match the number of granules for the other products. */
                if(aggr_num_granules != *exf_dat->ngranules) {
                    NAGG_ERROR("get_product_props():\n");
                    HDfprintf(stderr, "Example file %s has multiple products with different aggregate numbers.\n",
                            exf_dat->filepath);
                    HDfprintf(stderr, "Aggregate numbers must match for nagg to use this file.\n");
                    status = FAIL;
                    goto done;
                }
            } else {
                *exf_dat->ngranules = aggr_num_granules;
            }

            break;
        default:
            HDfprintf(stderr, "Object %s is not a group.\n", name);
        break;
    }

done:
    if (attr > -1)
        H5Aclose(attr);
    if (dset > -1)
        H5Dclose(dset);
    if (grpid > -1)
        H5Gclose(grpid);
    return status;
}


/* verify that *pt contains an integer.
 * Integer is defined as list of digits (0-9) with an optional negative sign ('-').
 * Return TRUE if it is an integer;
 * Otherwise FALSE.
 */
static int
HDisint(char *pt)
{
    if (!pt){
	return FALSE;
    }
    if (*pt == '-')
	pt++;
    while (HDisdigit(*pt)){
	if (!(*++pt))
	    break;
    }
    return (*pt ? FALSE : TRUE);
}

/* Purpose:
 *    Debug print the product list. Expect the first nlist elements are
 *    null-terminated DPIDs.
 */
static void
dprint_products_list(int nlist, char **plist)
{
    int i;
    assert(plist);
    assert(nlist>=0);
    printf("Debug printing property list:\n");
    for (i=0; i<nlist; i++){
	printf("plist[%02d]: %s\n", i, plist[i]);
    }
    printf("Done Debug printing property list\n");
}

/* Purpose:
 *    Parse command line options, validate the option values and set the option
 *    variables so that the tool may execute according to user request.
 * return: 0 if succeess;
 *    leave(EXIT_FAILURE) otherwise.
 */
int
parse_options(int argc, char * const argv[])
{
    int index;
    int i;
    int c;
    char *new_p;
    char *products_arg=NULL;
    char *exfile=NULL;		/* hold example file of like option */
    char *next_product;
    char *next_geoproduct=NULL;	/* hold previous located geoproduct */
    char *tmp_product=NULL;     /* hold product name if -A is used. */
    iet_t tmp_product_duration; /* duration time of tmp_product */
    int  aggregatetime=-1;	/* Aggregation time. -1 stands for not set. */
    int	 version_arg=0;		/* Version option. Init to no. */
    /* command line options: See function usage for a description */
    char *nagg_options = "A:d:D:g:hl:n:O:St:V";
#ifdef HAVE_GETOPT_H
    /* Use getopt_long if available */
    struct option nagg_long_options[] = {
	{"debug", 0, &debug_arg, 1},
	{"directory", 1, 0, 'd'},
	{"help", 0, 0, 'h'},
	{"like", 1, 0, 'l'},
	{"number", 1, 0, 'n'},
	{"simple", 0, 0, 'S'},
	{"type", 1, 0, 't'},
        {"version", 0, &version_arg, 1},
	{0, 0, 0, 0}
    };
    int long_option_index = 0;
#endif

    opterr = 0;

    /* Extract program name from the last component of called name */
    if((progname_g=HDstrrchr(argv[0], '/')))
	progname_g++;
    else
	progname_g = argv[0];

    /* initialize like_products_list to point at elements of like_products_list_array. */
    for (i=0; i<NAGG_Product_list_max; i++){
	like_products_list[i] =  &like_products_list_array[i][0];
    }
    
    while (1){
	c = 
#ifdef HAVE_GETOPT_LONG
	    getopt_long (argc, argv, nagg_options, nagg_long_options,
		&long_option_index);
#else
	    getopt (argc, argv, nagg_options);
#endif
	if (-1 == c)
	    break;
	switch (c) {
	  case 0:	/* one of the long options matched */
	    /* no more action needed yet. */
#ifdef DEBUG
	    printf("long option %s found\n",
		nagg_long_options[long_option_index].name);
#endif
	    break;
	  case 'h':
	    usage();
	    print_product_lists();
	    leave(EXIT_SUCCESS);
	    break;
	  case 'n':
	    if (!HDisint(optarg)){
		NAGG_ERROR("parse_options(): number of granules must be an integer\n");
		leave(EXIT_FAILURE);
	    }else{
		ngranulesperfile = atoi(optarg);
	    };
	    if (ngranulesperfile < 1){
		fprintf(stderr, "Option error: number of granules must be positive. Got %d\n",
		    ngranulesperfile);
		usage();
		leave(EXIT_FAILURE);
	    }
	    break;
	  case 'A':
	    if (!HDisint(optarg)){
		NAGG_ERROR("parse_options(): aggregation time must be an integer\n");
		leave(EXIT_FAILURE);
	    }else{
		aggregatetime = atoi(optarg);
	    };
	    if (aggregatetime < 1){
		fprintf(stderr, "Option error: aggregate time must be a positive integer. Got %d\n",
		    aggregatetime);
		usage();
		leave(EXIT_FAILURE);
	    }
	    break;
	  case 't':
	    /* dup the argument because it maybe modified later */
	    if (NULL==(products_arg=HDstrdup(optarg))){
		NAGG_ERROR("parse_options(): no memory for strdup\n");
		leave(EXIT_FAILURE);
	    }
	    break;
	  case 'l':
	    exfile = optarg;
	    if (get_example_props(exfile, &like_ngranules, &like_nproducts, like_products_list, &like_ispackaged) < 0){
		NAGG_ERROR("like example file encountered failure.\n");
		leave(EXIT_FAILURE);
	    }
#ifdef DEBUG
	    printf("get_example_props returns like_ngranules=%d, like_nproducts=%d\n", like_ngranules, like_nproducts);
            printf("                          geoproduct=%s, like_ispackaged=%d\n", geoproduct, like_ispackaged);
	    dprint_products_list(like_nproducts, like_products_list);
#endif
	    /* verify returned values from get_example_props */
          /* LRK 20130225  These assertions may not hold for extended like 
             option capabilities.  If they don't it might be necessary to 
             move them and check them in combination with other values. */  
	  /*  assert(like_ngranules > 0); */
	  /*  assert(like_nproducts > 0); */
            if (geoproduct && like_nproducts <= 0) 
                geofiles_arg=GEOFILE_GEOPRODUCT;
            if (!like_ispackaged) 
		outfileformat = UNPACKAGED;
            /* like_ispackaged==1 if the example file has no N_GEO_Ref attribute,
               and geoproduct is NULL if the example file has no geoproduct group.
               Those 2 conditions indicate no geolocation should be produced. */
            if (like_ispackaged && !geoproduct)
                geofiles_arg=GEOFILE_NO;
	    break;
	  case 'd':
	    outDir = optarg;
	    if (!isdirectory(outDir)){
		NAGG_ERROR("output directory must be existing.\n");
		leave(EXIT_FAILURE);
	    }
	    break;
	  case 'O':
	    origin_arg = optarg;
	    break;
	  case 'D':
	    domain_arg = optarg;
	    break;
	  case 'g':
	    if (HDstreq(optarg,"yes")||HDstreq(optarg,"1")){
		geofiles_arg=GEOFILE_YES;
	    }else if (HDstreq(optarg,"no")||HDstreq(optarg,"0")){
		geofiles_arg=GEOFILE_NO;
	    }else if (HDstreq(optarg,"strict")||HDstreq(optarg,"2")){
		geofiles_arg=GEOFILE_STRICT;
	    }else{
		/* check if it is a Geoproduct ID */
		if (NULL!=get_product_sname_by_id(optarg)){
		    geoproduct=optarg;
		    geofiles_arg=GEOFILE_GEOPRODUCT;
		}else{
		    NAGG_ERROR("parse_options(): illegal geolocation criterion given\n");
		    usage();
		    leave(EXIT_FAILURE);
		}
	    }
	    break;
	  case 'S':
		outfileformat = UNPACKAGED;
		break;
	  case '?':
	    if (optopt == 'c')
	      fprintf (stderr, "Option -%c requires an argument.\n", optopt);
	    else if (isprint (optopt))
	      fprintf (stderr, "Unknown option `-%c'.\n", optopt);
	    else
	      fprintf (stderr,
		"Unknown option character `\\x%x'.\n",
		optopt);
	    usage();
	    leave(EXIT_FAILURE);
	  default:
	    NAGG_ERROR("parse_options(): getopt returned unexpected value.\n");
	    fprintf(stderr, "Unexpected value is %d\n", c);
	    leave(EXIT_FAILURE);
	}
    }

    /* Get the input file arguments, assume remainings are all input files. */
    ninputfiles = argc-optind;
    inputfiles = argv+optind;

    /* Show version information if requested. */
    if (version_arg){
        print_nagg_version();
	/* exit if there is no input files given. */
	if (ninputfiles <= 0)
	    leave(EXIT_SUCCESS);
    }

    /* If like option is used, assigned its values from the example file
     * to the various options (-n and -t ) if they are not given.
     */
    if (exfile){
	/* is neither -n nor -A is given? */
	if (-1 == ngranulesperfile && -1 == aggregatetime){
	    ngranulesperfile = like_ngranules;
	}
	/* is -t not given? */
	if (NULL==products_arg){
	    nproducts=like_nproducts;
	    products_list=like_products_list;
	    tmp_product = products_list[0];	/* Need it if -A is used. */
	}
    }
    
    /* If -t is given, copy the comma separate argument list to the products_list */
    if (products_arg){
	nproducts=0;
	products_list=&products_list_array[0];
	{
	    /* We have a list of products separated by commas. Scan down the list,
	     * replacing commas by \0 and assign each to the products_list_array.
	     */
	    products_list[0] = products_arg;
	    next_product = products_arg;
	    tmp_product = products_list[0];	/* Need it if -A is used. */
	    do{
		if (nproducts >= NAGG_Product_list_max ){
		    NAGG_ERROR("parse_options(): too many products requested.\n");
		    fprintf(stderr, "Maximum number of products is %d\n", NAGG_Product_list_max);
		    leave(EXIT_FAILURE);
		}
		new_p = HDstrchr(next_product, ',');
		if (new_p){
		    /* replace the comma with a null terminator */
		    *new_p = '\0';
		}
		if (NULL==get_product_sname_by_id(next_product)){
		    NAGG_ERROR("parse_options(): illegal product ID given:\n");
                    fprintf(stderr, "%s\n", next_product);
		    usage();
		    leave(EXIT_FAILURE);
		}
		products_list[nproducts++] = next_product;
		if (new_p){
		    next_product = new_p+1;
		}
	    }while (new_p);
	}
    }

    /* If the product list is empty, there must be a geoproduct for geo only
       aggregation.  Otherwise nagg has nothing to process. */
    if (NULL==products_list){
	if (geofiles_arg!=GEOFILE_GEOPRODUCT){
	    NAGG_ERROR("parse_options(): no product type ID given\n");
	    usage();
	    leave(EXIT_FAILURE);
	}
	tmp_product = geoproduct;	/* Need it if -A is used. */
    }else{
        /* check the lists of products and geoproduct if given with -g for compatibility,
           and if there are compatible geoproducts, determine which one to process. */
        set_geoproduct_to_process(products_list);


	/* Find the Geoproduct used unless -g no is given */
        /* 20130125 If both -t/-l and -g <GEOproduct> were given, use <Geoproduct>. */




    }
    /* if the -g argument was no or 0, we used the geoproduct to check for compatibility,
        but we need to set it back to NULL or nagg_get_granules will try to process the 
        geoproduct.  Maybe there's a better way? */
    if(geofiles_arg==GEOFILE_NO) {
        geoproduct=NULL;
    }
    
    /* Check Origin name */
    if (NULL==origin_arg){
	origin_arg="XXXX";
    };
    /* Check Domain name */
    if (NULL==domain_arg){
	domain_arg="XXX";
    };
    /* check if at least one input file is provided unless --version is used. */
    if (ninputfiles <= 0){
	NAGG_ERROR("parse_options(): no input file given\n");
	usage();
	leave(EXIT_FAILURE);
    };
    /* Check if -A is used, that -n is not used, then calculate
     * ngranulesperfile from aggregatetime.
     */
    if (aggregatetime != -1){
	if (ngranulesperfile != -1) {
	    NAGG_ERROR("Only one of -A or -n may be used\n");
	    usage();
	    leave(EXIT_FAILURE);
	}
    assert(tmp_product);
    if (0 == (tmp_product_duration=get_product_duration_by_id(tmp_product))){
	/* product duration time not found */
	NAGG_ERROR("parse_options(): can't find duration time\n");
	usage();
	leave(EXIT_FAILURE);
    };
    ngranulesperfile = (int)ceilf(aggregatetime/micros2s(tmp_product_duration));
#ifdef DEBUG
    printf("aggregatetime=%d, ngranulesperfile=%d, tmp_product_duration=%ull\n",
	aggregatetime, ngranulesperfile, tmp_product_duration);
#endif
    }else{
	/* if -n is not given, set ngranulesperfile to default value of 1 */
	if (ngranulesperfile == -1)
	    ngranulesperfile = 1;
    };
    /* Check that ngranulesperfile has valid value */
    if (ngranulesperfile < 1){
	NAGG_ERROR("parse_options(): number of granules is illegal\n");
	fprintf(stderr, "ngranulesperfile=%d\n", ngranulesperfile);
	leave(EXIT_FAILURE);
    }

    /* All done. */
    return(0);
}


/*
 * Function: set_geoproduct_to_process
 *
 * Purpose:   1) Set geoproduct as follows:
 *                A. geoproduct = <geoproduct> from -g <geoproduct> 
 *                   if that command option was used.
 *                B. geoproduct = Geoproduct(products_list[0]) 
 *                   if all products in products_list have the same
 *                   designated geoproduct.
 *                C. geoproduct = terrain corrected geoproduct if 
 *                   some products have not terrain corrected GEO product 
 *                   designated and other products have the corresponding
 *                   terrain corrected GEO product designated.
 *                     
 *            2) Check products and geoproducts for compatibility
 *
 * Parameter: IN products_list: a list of NPP data products to be processed
 *
 * Note:       geoproduct_to_process is set to the specified geoproduct if it was 
 *             set by -g <geoproduct>.  In this case, this function checks for 
 *             compatibility of products, but sets geoproduct to its original value 
 *             before returning (except when exiting the program for errors).  If 
 *             geoproduct was not set it is NULL, and geoproduct will be set to the 
 *             geoproduct from the products table for the first product in the list, 
 *             or to the terrain corrected product if there are compatible products 
 *             with both not terrain corrected and related terrain corrected 
 *             geolocation products.
 *
 * Return:     Success:     0
 *             Failure:     -1
 *
 * Programmer: Larry Knox
 *             February 27, 2013
 */
int 
set_geoproduct_to_process(char **products_list)
{
    int i, status = 0;
    char *next_product;
    char *next_geoproduct=NULL;	/* hold previous located geoproduct */
    char *geoproduct_to_process = geoproduct; 

    /* Always check products and geoproduct for compatibility. */
    if (geoproduct==NULL){
        geoproduct = get_gpid_by_id(products_list[0]);
    }
    for (i=0; i<nproducts; i++){
                
        next_product = products_list[i];
        if (NULL==(next_geoproduct=get_gpid_by_id(next_product))){
            NAGG_ERROR("parse_options(): could not find Geolocation product ID\n");
            usage();
            leave(EXIT_FAILURE);
        }
        /* See if it is different from previous found geoproduct */
        if (HDstrneq(next_geoproduct, geoproduct)) {
            /* If the pair geoproduct, next_geoproduct is found in equiv_geoprod_table, the second entry is tcgeoproduct and should be assigned to geoproduct_to_process if it has not already been set.  */
            if (is_pair_geoproduct_and_tcgeoproduct(next_geoproduct, geoproduct)) {
                if(!geoproduct_to_process) {
                    geoproduct_to_process = geoproduct;
                }
            } else if (is_pair_geoproduct_and_tcgeoproduct(geoproduct, next_geoproduct)) {
                if(!geoproduct_to_process) {
                    geoproduct_to_process = next_geoproduct;
                }
            } else {
                NAGG_ERROR("parse_options(): found incompatible product ID:\n");
                fprintf(stderr, "%s\n", next_product);
                usage();
                leave(EXIT_FAILURE);
            }
        }
    }
    if (geoproduct_to_process != NULL) 
        geoproduct = geoproduct_to_process;
    
    return status;
}

/*
 * Function: is_pair_geoproduct_and_tcgeoproduct
 *
 * Purpose:   Check 2 geoproducts to see if the second is the terrain 
 *            corrected version of the first, according to the 
 *            equiv_geoprod_table (get_tcpid_by_gpid()).
 *                     
 * Parameter: IN geoproduct1, geoproduct2 
 *
 * Return:     1 if geoproduct2 is the terrain corrected version of geoproduct1
 *             0 otherwise
 *
 * Programmer: Larry Knox
 *             February 27, 2013
 */
int
is_pair_geoproduct_and_tcgeoproduct(char *geoproduct1, char *geoproduct2) 
{
    char *tceq_id = NULL;
    if(HDstrneq(geoproduct1, geoproduct2)) {
        /* Check if geoproduct2 is terrain corrected version of geoproduct1 */
        tceq_id = get_tcpid_by_gpid(geoproduct1);
        if(tceq_id && HDstreq(geoproduct2, tceq_id))
            return 1;
    }
    return 0;
}
